from enum import Enum
from typing import Optional
import base64

from pydantic import BaseModel, model_validator


class Singelton:
    _instances = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            instance = super(Singelton, cls).__new__(cls)
            cls._instances[cls] = instance
        return cls._instances[cls]


class DestinationType(str, Enum):
    HTTP = "HTTP"


class ProxyType(str, Enum):
    INTERNET = "Internet"
    ONPREMISE = "OnPremise"


class AuthType(str, Enum):
    NO_AUTH = "NoAuthentication"
    BASIC = "BasicAuthentication"
    OAUTH2_CLIENT_CREDENTIALS = "OAuth2ClientCredentials"
    PRINCIPAL_PROPAGATION = "PrincipalPropagation"


class EnvProxyType(str, Enum):
    INTERNET = "Internet"
    ONPREMISE = "OnPremise"


class EnvAuthType(str, Enum):
    NO_AUTH = "NoAuthentication"
    BASIC = "BasicAuthentication"


class EnvDestination(BaseModel):
    name: str
    url: str
    type: DestinationType
    proxy_type: EnvProxyType
    auntentication: EnvAuthType
    username: Optional[str] = None
    password: Optional[str] = None
    forwardAuthToken: Optional[bool] = None
    proxyHost: Optional[str] = None
    proxyPort: Optional[int] = None
    timeout: Optional[int] = None

    @model_validator(mode="after")
    def validate_auth_fields(self):
        if self.auntentication == EnvAuthType.BASIC:
            if not self.username or not self.password:
                raise ValueError(
                    "Username and password are required for Basic Authentication."
                )
        return self

    @model_validator(mode="after")
    def validate_proxy(self):
        if self.proxy_type == EnvProxyType.ONPREMISE:
            if not self.proxyHost and not self.proxyPort:
                raise ValueError(
                    "Proxy Host and Port are required for OnPremise proxy type."
                )
        return self

    def get_configuration(self) -> dict:
        config = {
            "destinationConfiguration": {
                "Name": self.name,
                "Type": self.type.value,
                "URL": self.url,
                "ProxyType": self.proxy_type.value,
                "Authentication": self.auntentication.value,
                "ForwardAuthToken": self.forwardAuthToken,
                "Timeout": self.timeout,
            },
        }

        if self.auntentication == EnvAuthType.BASIC:
            config["authToken"] = [self._get_auth_token()]

        if self.proxy_type == EnvProxyType.ONPREMISE:
            config["proxyConfiguration"] = (
                {"proxy": f"{self.proxyHost}:{self.proxyPort}"}
                if self.proxyHost and self.proxyPort
                else {}
            )

    def _get_auth_token(self) -> dict:
        token_str = f"{self.username}:{self.password}"
        token_bytes = token_str.encode("utf-8")
        base64_bytes = base64.b64encode(token_bytes)
        base64_token = base64_bytes.decode("utf-8")
        return {
            "type": "Basic",
            "value": base64_token,
            "http_header": {"key": "Authorization", "value": f"Basic {base64_token}"},
        }



from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic import model_validator
from .models import EnvDestination


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8")

    vcap_services: dict | None = None

    destinations: list[EnvDestination] | None = None

    @model_validator(mode="after")
    def validate_vcap_variables(self):
        vcap = self.vcap_services.get("destination") if self.vcap_services else None
        if not self.destinations and not vcap:
            raise ValueError("Destination service not bound to the application")

        return self



from datetime import datetime, timedelta

import httpx

from .settings import Settings
from .connectivity import Connectivity
from .models import Singelton, ProxyType


class DestinationInstance(Singelton):
    settings = Settings()
    auth: dict = {"token": None, "expires_at": None}

    def __init__(self):
        self._init_instance_credentials()
        self.local_destination = self.settings.destinations

    def _init_instance_credentials(self):
        """
        Initialize instance credentials from settings.
        """
        if self.settings.vcap_services and self.settings.vcap_services.get(
            "destination"
        ):
            vcap = self.settings.vcap_services.get("destination")  # type: ignore
            credentials = vcap[0].get("credentials")  # type: ignore

            self.base_url = credentials.get("uri")
            self.client_id = credentials.get("clientid")
            self.client_secret = credentials.get("clientsecret")
            self.token_service_url = credentials.get("url")

    async def _generate_auth_token(self) -> str:
        """
        Generate an authentication token using client credentials.

        Returns:
            None
        """

        # Return cached token if valid
        if self.auth["token"] is not None and self.auth["expires_at"] > datetime.now():
            return self.auth["token"]
        # Generate new token
        else:
            try:
                async with httpx.AsyncClient(
                    base_url=self.token_service_url,
                    auth=httpx.BasicAuth(self.client_id, self.client_secret),
                ) as client:
                    response = await client.post(
                        "/oauth/token",
                        params={
                            "grant_type": "client_credentials",
                        },
                    )
                    response.raise_for_status()
                    token_data = response.json()
                    self.auth["token"] = token_data.get("access_token")
                    self.auth["expires_at"] = datetime.now() + timedelta(
                        seconds=token_data.get("expires_in", 3600)
                    )
            except httpx.HTTPError as e:
                raise ConnectionError(
                    f"Error generating auth token for destination instance: {e}"
                ) from e

    async def get_configuration(self, name: str) -> dict:
        """
        Retrieve the configuration for a specific destination.

        Args:
            name (str): The name of the destination.
        Returns:
            dict: The destination configuration.
        """
        # Check if destination exists in environment variables
        if self.local_destination is not None:
            for dest in self.local_destination:
                if dest.name == name:
                    return dest.get_configuration()
        # Fetch destination configuration from the service
        else:
            await self._generate_auth_token()
            headers = {
                "Authorization": f"Bearer {self.auth['token']}",
                "Content-Type": "application/json",
            }

            try:
                async with httpx.AsyncClient(
                    base_url=self.base_url, headers=headers
                ) as client:
                    response = await client.get(
                        f"/destination-configuration/v1/destinations/{name}"
                    )
                    response.raise_for_status()
                    result = response.json()

                    proxy_type = result["destinationConfiguration"].get("ProxyType")

                    # Add proxy configuration if proxy type is OnPremise
                    if proxy_type == ProxyType.ONPREMISE.value:
                        connectivity = Connectivity()

                        result["proxyConfiguration"] = {
                            "proxy": await connectivity.get_proxy_config()
                        }
                    return result
            except httpx.HTTPError as e:
                raise ConnectionError(
                    f"Error fetching destination configuration for '{name}': {e}"
                ) from e



import httpx
from datetime import datetime, timedelta

from .models import Singelton
from .settings import Settings


class Connectivity(Singelton):
    settings = Settings()
    auth: dict = {"token": None, "expires_at": None}

    def __init__(self) -> None:
        self._init_instance_credentials()

    def _init_instance_credentials(self) -> None:
        """
        Initialize instance credentials from settings.
        """
        if self.settings.vcap_services and self.settings.vcap_services.get(
            "connectivity"
        ):
            vcap = self.settings.vcap_services.get("connectivity")  # type: ignore
            credentials = vcap[0].get("credentials")  # type: ignore

            self.token_url = credentials.get("url")
            self.client_id = credentials.get("clientid")
            self.client_secret = credentials.get("clientsecret")
            self.onpremise_proxy_host = credentials.get("onpremise_proxy_host")
            self.onpremise_proxy_port = credentials.get("onpremise_proxy_port")
        else:
            raise ValueError("Connectivity service not bound to the application")

    async def _get_auth_token(self) -> None:
        """
        Generate an authentication token using client credentials.

        Returns:
            None
        """

        # Return cached token if valid
        if self.auth["token"] is not None and self.auth["expires_at"] > datetime.now():
            return self.auth["token"]
        # Generate new token
        else:
            try:
                async with httpx.AsyncClient(
                    base_url=self.token_url,
                    auth=httpx.BasicAuth(self.client_id, self.client_secret),
                ) as client:
                    response = await client.post(
                        "/oauth/token",
                        params={"grant_type": "client_credentials"},
                    )
                    response.raise_for_status()
                    token_data = response.json()

                    # Cache the token with its expiry time
                    self.auth["token"] = token_data.get("access_token")
                    self.auth["expires_at"] = datetime.now() + timedelta(
                        seconds=token_data.get("expires_in", 3600)
                    )

            except httpx.HTTPError as e:
                raise ConnectionError(f"Failed to obtain auth token: {e}") from e

    async def get_proxy_config(self) -> dict:
        """
        Get the OnPremise proxy configuration.

        Returns:
            dict: Proxy configuration with host and port.
        """
        await self._get_auth_token()

        return {
            "proxy": f"http://{self.onpremise_proxy_host}:{self.onpremise_proxy_port}",
            "headers": {"Proxy-Authorization": f"Bearer {self.auth['token']}"},
        }



import httpx
import asyncio
from datetime import datetime, timedelta

from .core import DestinationInstance, ProxyType, AuthType


class Destination:
    """
    SAP BTP Destination client
    """

    _instance = DestinationInstance()
    _config: dict = {"parameters": None, "expires_in": None}
    auth_type: str = None
    proxy_type: str = None
    forward_auth_token: bool = False
    url: str = None
    timeout: int = 30

    def __init__(self, name: str, cache_duration: int = 3600) -> None:
        if name is None or name.strip() == "":
            raise ValueError("Destination name must be provided and cannot be empty.")
        else:
            self.name = name

        self.cache_duration = cache_duration

    def _validate_configuration(self, config: dict) -> None:
        """
        Validate the retrieved destination configuration.
        """
        proxy_type = config.get("destinationConfiguration").get("ProxyType")
        auth_type = config.get("destinationConfiguration").get(
            "Authentication"
        )

        if proxy_type not in ProxyType:
            raise ValueError(f"Unsupported proxy type: {proxy_type}")

        if auth_type not in AuthType:
            raise ValueError(f"Unsupported authentication type: {auth_type}")

    async def _get_configuration(self) -> dict:
        """
        Retrieve the destination configuration.

        Returns:
            dict: The destination configuration.
        """
        config = await self._instance.get_configuration(self.name)

        destinationConfig = config.get("destinationConfiguration")

        self.auth_type = destinationConfig.get("Authentication")
        self.proxy_type = destinationConfig.get("ProxyType")
        self.forward_auth_token = destinationConfig.get("ForwardAuthToken", False)
        self.url = destinationConfig.get("URL")
        self.timeout = destinationConfig.get("Timeout", 30)

        self._validate_configuration(config)

        return config

    async def _build_parameters(self, jwt: str) -> None:
        """
        Build the parameters dictionary from the configuration.
        """
        if (
            self.auth_type is not None and self.auth_type != AuthType.OAUTH2_CLIENT_CREDENTIALS.value
            and self._config["parameters"] is not None
        ):
            return self._config["parameters"]
        elif (
            self.auth_type is not None and self.auth_type == AuthType.OAUTH2_CLIENT_CREDENTIALS.value
            and self._config["expires_in"] is not None
            and self._config["expires_in"] > datetime.now()
        ):
            return self._config["parameters"]
        else:
            config = await self._get_configuration()

            params = {
                "base_url": self.url,
                "timeout": self.timeout,
            }

            if (
                config.get("authTokens") is not None
                and len(config.get("authTokens")) > 0
            ):
                params["headers"] = {
                    config["authTokens"][0]["http_header"]["key"]: config["authTokens"][
                        0
                    ]["http_header"]["value"]
                }
                self._config["expires_in"] = datetime.now() + timedelta(
                    seconds=int(config["authTokens"][0].get(
                        "expires_in", self.cache_duration
                    ))
                )

            if self.auth_type == AuthType.NO_AUTH.value and self.forward_auth_token:
                if jwt is None:
                    raise ValueError("JWT must be provided for forwarding auth token.")
                else:
                    if "headers" not in params:
                        params["headers"] = {}
                    params["headers"]["Authorization"] = f"Bearer {jwt}"

            if self.proxy_type == ProxyType.ONPREMISE.value:
                proxy_config = config.get("proxyConfiguration", {})
                if proxy_config.get("proxy") is not None:
                    params["proxy"] = proxy_config.get("proxy")
                    params["headers"] = {
                        **params.get("headers", {}),
                        **proxy_config.get("headers", {}),
                    }

                if self.auth_type == AuthType.PRINCIPAL_PROPAGATION.value:
                    if jwt is None:
                        raise ValueError(
                            "JWT must be provided for Principal Propagation."
                        )
                    else:
                        params["headers"]["SAP-Connectivity-Authentication"] = (
                            f"Bearer {jwt}"
                        )

                if (
                    config["destinationConfiguration"].get("CloudConnectorLocationId")
                    is not None
                ):
                    params["headers"]["SAP-Connectivity-SCC-Location_ID"] = config[
                        "destinationConfiguration"
                    ].get("CloudConnectorLocationId")

            self._config["parameters"] = params

    def get_client(self, jwt: str = None) -> httpx.Client:
        """
        Create and return an HTTP client configured for the destination.

        Returns:
            httpx.Client: An HTTP client instance.
        """
        asyncio.run(self._build_parameters(jwt))
        return httpx.Client(**self._config["parameters"])

    async def get_aclient(self, jwt: str = None) -> httpx.AsyncClient:
        """
        Create and return an asynchronous HTTP client configured for the destination.

        Returns:
            httpx.AsyncClient: An asynchronous HTTP client instance.
        """
        await self._build_parameters(jwt)
        return httpx.AsyncClient(**self._config["parameters"])
